from vcPythonProcessHandler import *

app = getApplication()
sim = getSimulation()
comp = getComponent()

#Initialize failure manager and routine property
def OnStart():
  global failure_mgr, is_broken
  routine = exe.Processes[0]
  is_broken = routine.getProperty('IsBroken')
  if not is_broken:
    is_broken = routine.createProperty(VC_BOOLEAN, 'IsBroken')
  failure_mgr = FailureManager('Failure::MTBF', 'Failure::MTTR', 'Failure::FailureMode',
  'Failure::AutomaticRepair', 'Failure::CyclesBeforeFailure', 'Failure::CreateRepairStatements')



  
  
#Executed by the process executor
########################
def OnStatementExecute(executor, statement):
  ###################
  
  check_failure()
  if is_broken.Value:
    return
  if comp.SimulationLevel == VC_SIMULATION_FAST:
    #delay(time)
    check_failure()
    failure_mgr.current_cycle += 1
    return


####PROCESS HERE####

  check_failure()
  failure_mgr.current_cycle += 1


#Failure related methods

def check_failure():
  #The machine is alread broken. Only when auto repair is disabled
  if is_broken.Value:
    failure_mgr.reset()
  #The machine is not yet broken. Evaluate if the machine should break now
  else:
    failed = failure_mgr.evaluate_failure()
    if failed:
      stats.State = 'Broken'
      is_broken.Value = failed
      #The Machine is repaired automatically after the delay if auto repair is enabled
      #If auto repair is disabled user must add GetProperty and If statements to call human to do the maintenance
      if auto_repair.Value:
        delay(failure_mgr.repair_time)
        failure_mgr.reset()
        stats.State = 'Busy'
  return is_broken.Value


class FailureManager(object):
  # Call evaluate_failure method frequently to check for failures and reset failure by calling reset.
  # Will create new mtbf and mttr properties if not existing
  def __init__(self, mtbf_prop_name, mttr_prop_name, mode_prop_name, auto_repair_prop_name, cycles_prop_name, button_name):
    # type: (str, str) -> None
    self.fail_mode_prop = self._validate_property(comp, VC_STRING, mode_prop_name, default_val = 'MTBF', constraint = VC_PROPERTY_STEP, step_vals = ['MTBF', 'Cycles before failure'])
    self.mtbf_prop = self._validate_property(comp, VC_DISTRIBUTION, mtbf_prop_name, unit='Time')
    self.cycles_prop = self._validate_property(comp, VC_DISTRIBUTION, cycles_prop_name, visible = False)
    self.mttr_prop = self._validate_property(comp, VC_DISTRIBUTION, mttr_prop_name, unit='Time')
    self.auto_rep_prop = self._validate_property(comp, VC_BOOLEAN, auto_repair_prop_name, default_val = True, visible = False) #VISIBLE
    self.create_stmnts_prop = self._validate_property(comp, VC_BUTTON, button_name, visible = False)
    self.current_cycle = 0
    self.reset()
 
  def evaluate_failure(self):
    # type: () -> bool
    # Returns True on failure, False othwerwise. Sets also failure_active property to True
    #MTBF is used
    if self.fail_mode_prop.Value == 'MTBF':
      if 0.0 < self.next_failure_time <= sim.SimTime:
        self.failure_active = True
    #MTBF not used
    else:
      if self.current_cycle >= self.failing_cycle:
        self.failure_active = True
    return self.failure_active

  def reset(self):
    # type: () -> None
    # Must be called after repair is complete
    # Ends failure condition and retrieves next failure values
    self.failure_active = False
    is_broken.Value = False
    if self.fail_mode_prop.Value == 'MTBF':
      if self.mtbf_prop.Value > 0.0:
        self.next_failure_time = sim.SimTime + self.mtbf_prop.Value
      else:
        self.next_failure_time = 0.0
    else:
      self.failing_cycle = round(self.cycles_prop.Value)
    self.repair_time = self.mttr_prop.Value
    #Prevent negative repair time
    while self.repair_time < 0:
      self.repair_time = self.mttr_prop.Value
    self.current_cycle = 0

  def _validate_property(self, component, prop_type, prop_name, default_val=None, unit=None, visible=True, constraint = None, step_vals = None):
    # type: (vcComponent, str, str, str, bool) -> vcProperty
    prop = component.getProperty(prop_name)
    if prop and prop.Type != prop_type:
      prop = component.deleteProperty(prop)
    if not prop:
      if constraint:
        prop = component.createProperty(prop_type, prop_name, constraint)
      else:
        prop = component.createProperty(prop_type, prop_name)
      if step_vals:
        prop.StepValues = step_vals
      if default_val is not None:
        prop.Value = default_val
      prop.IsVisible = visible
      if unit:
        prop.Quantity = unit
      prop.WritableWhenSimulating = False
    return prop

#Add default statements for manual repair
def add_statements(prop):
  routine = exe.Processes[0]
  statements = routine.Statements
  #Add If Statement after the custom statement
  index = 0
  for s in statements:
    if s.Type == VC_STATEMENT_PROCESS:
      break
    index += 1
  #Check if there are old repair statements and delete the old ones
  next_statement = routine.Statements[index+1]
  if next_statement:
    if next_statement.Type == VC_STATEMENT_PROCESSIF:
      if next_statement.Condition == 'IsBroken':
        routine.deleteStatement(next_statement)
  if_statement = routine.addStatement(VC_STATEMENT_PROCESSIF, index + 1)
  if_statement.Condition = 'IsBroken'
  if_scope = if_statement.ThenScope
  #Add Statements to the If scope
  #Work, call human
  work_statement = if_scope.addStatement(VC_STATEMENT_WORK, -1)
  #Find  human TC and assign to Controller
  transport_node = comp.findBehaviour('TransportNode')
  for link in transport_node.TransportLinks:
    implementer = link.Implementer
    if implementer:
      if implementer.Type == VC_PYTHONTRANSPORTCONTROLLER:
        if implementer.Icon == 'UPM/pHumanTransportController':
          human_tc = implementer.Component
          break
  else:
    human_tc = app.findComponent('Human Transport Controller')
  tc = None
  if human_tc:
    tc = human_tc.findBehaviour('TC')
    if tc:
      work_statement.Controller = tc
  work_statement.ProcessTime = 0.0
  work_statement.ResourcePositionFrame = 'SetupLocation'
  work_statement.WorkPositionFrame = 'SetupLocation'
  #Set state to Repair
  state_statement = if_scope.addStatement(VC_STATEMENT_SETSTATISTICSSTATE , -1)
  state_statement.Statistics = stats
  state_statement.State = 'Repair'
  #Work, Start Repairing
  work_statement = if_scope.addStatement(VC_STATEMENT_WORK , -1)
  if tc:
    work_statement.Controller = tc
  #work_statement.ProcessTime = mttr_prop.Value
  work_statement.TimeSource = VC_TIMESOURCE_EXPRESSION
  work_statement.ProcessTimeExpression = 'Failure::MTTR'
  work_statement.ResourcePositionFrame = 'SetupLocation'
  work_statement.WorkPositionFrame = 'SetupLocation'
  #Set state to Busy
  state_statement = if_scope.addStatement(VC_STATEMENT_SETSTATISTICSSTATE , -1)
  state_statement.Statistics = stats
  state_statement.State = 'Busy'
  #Run the machine process again once the machine is repaired
  custom_statement = if_scope.addStatement(VC_STATEMENT_PROCESS , -1)
  custom_statement.getProperty('Tool').IsVisible = False
  custom_statement.getProperty('Base').IsVisible = False
  custom_statement.Description = 'Custom Machine Process'
  custom_statement.Process = python_process_handler
  custom_statement.Name = 'Failed Machine Process'
  print 'Statements for repairing the machine were added to the routine. Set the statement parameters before running the simulation.'


def prop_changed (prop):
  if prop.Name == fail_mode.Name:
    if prop.Value == 'MTBF':
      toggle_prop_vis(mtbf_prop, True)
      toggle_prop_vis(cycles_prop, False)
    else:
      toggle_prop_vis(mtbf_prop, False)
      toggle_prop_vis(cycles_prop, True)
  elif prop.Name == auto_repair.Name:
    toggle_prop_vis(create_statements_prop, not prop.Value)


def toggle_prop_vis(prop, visible):
  if visible:
    prop.IsVisible = True
  else:
    prop.IsVisible = False


mtbf_prop = comp.getProperty('Failure::MTBF')
mttr_prop = comp.getProperty('Failure::MTTR')
cycles_prop = comp.getProperty('Failure::CyclesBeforeFailure')
auto_repair = comp.getProperty('Failure::AutomaticRepair')
if auto_repair:
  auto_repair.OnChanged = prop_changed
fail_mode = comp.getProperty('Failure::FailureMode')
if fail_mode:
  fail_mode.OnChanged = prop_changed
create_statements_prop = comp.getProperty('Failure::CreateRepairStatements')
if create_statements_prop:
  create_statements_prop.OnChanged = add_statements
#time_prop = comp.getProperty('ProcessTime')

stats = comp.findBehaviour('Statistics')
python_process_handler = comp.findBehaviour('FailureHandler')
exe = comp.getBehaviour('ProcessExecutor__HIDE__')